@@ -1,7 +1,8 @@
+{-# LANGUAGE CPP #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.LargeWord
--- Copyright   :  (c) Dominic Steinitz 2004
+-- Copyright   :  (c) Dominic Steinitz 2004 - 2011
 -- License     :  BSD
 --
 -- Maintainer  :  dominic@steinitz.org
@@ -36,6 +37,7 @@ class LargeWord a where
    largeWordToInteger :: a -> Integer
    integerToLargeWord :: Integer -> a
    largeWordPlus :: a -> a -> a
+   largeWordMinus :: a -> a -> a
    largeWordAnd :: a -> a -> a
    largeWordOr :: a -> a -> a
    largeWordShift :: a -> Int -> a
@@ -48,6 +50,7 @@ instance LargeWord Word8 where
   largeWordToInteger = toInteger
   integerToLargeWord = fromInteger
   largeWordPlus = (+)
+  largeWordMinus = (-)
   largeWordAnd  = (.&.)
   largeWordOr   = (.|.)
   largeWordShift = shift
@@ -60,6 +63,7 @@ instance LargeWord Word16 where
   largeWordToInteger = toInteger
   integerToLargeWord = fromInteger
   largeWordPlus = (+)
+  largeWordMinus = (-)
   largeWordAnd  = (.&.)
   largeWordOr   = (.|.)
   largeWordShift = shift
@@ -69,26 +73,28 @@ instance LargeWord Word16 where
 -- Word32 is a key in the obvious way.
 
 instance LargeWord Word32 where
-   largeWordToInteger = toInteger
-   integerToLargeWord = fromInteger
-   largeWordPlus = (+)
-   largeWordAnd = (.&.)
-   largeWordOr = (.|.)
-   largeWordShift = shift
-   largeWordXor = xor
-   largeBitSize = bitSize
+  largeWordToInteger = toInteger
+  integerToLargeWord = fromInteger
+  largeWordPlus = (+)
+  largeWordMinus = (-)
+  largeWordAnd = (.&.)
+  largeWordOr = (.|.)
+  largeWordShift = shift
+  largeWordXor = xor
+  largeBitSize = bitSize
 
 -- Word64 is a key in the obvious way.
 
 instance LargeWord Word64 where
-   largeWordToInteger = toInteger
-   integerToLargeWord = fromInteger
-   largeWordPlus = (+)
-   largeWordAnd = (.&.)
-   largeWordOr = (.|.)
-   largeWordShift = shift
-   largeWordXor = xor
-   largeBitSize = bitSize
+  largeWordToInteger = toInteger
+  integerToLargeWord = fromInteger
+  largeWordPlus = (+)
+  largeWordMinus = (-)
+  largeWordAnd = (.&.)
+  largeWordOr = (.|.)
+  largeWordShift = shift
+  largeWordXor = xor
+  largeBitSize = bitSize
 
 -- Define larger keys from smaller ones.
 
@@ -100,7 +106,7 @@ loHalf (LargeKey a b) = a
 {-# INLINE hiHalf #-}
 hiHalf (LargeKey a b) = b
 
-instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) =>
+instance (Ord a, Bits a, Num a, LargeWord a, Bits b, Num b, LargeWord b) =>
    LargeWord (LargeKey a b) where
       largeWordToInteger (LargeKey lo hi) =
          largeWordToInteger lo + (2^(bitSize lo)) * largeWordToInteger hi
@@ -112,6 +118,10 @@ instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) =>
          LargeKey lo' hi' where
             lo' = alo + blo
             hi' = ahi + bhi + if lo' < alo then 1 else 0
+      largeWordMinus (LargeKey alo ahi) (LargeKey blo bhi) =
+         LargeKey lo' hi' where
+            lo' = alo - blo
+            hi' = ahi - bhi - if lo' > alo then 1 else 0
       largeWordAnd (LargeKey alo ahi) (LargeKey blo bhi) =
          LargeKey lo' hi' where
             lo' = alo .&. blo
@@ -126,31 +136,51 @@ instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) =>
             hi' = ahi `xor` bhi
       largeWordShift w 0 = w
       largeWordShift (LargeKey lo hi) x =
-         if bitSize lo < bitSize hi
-            then LargeKey (shift lo x) 
-                          (shift hi x .|. (shift (conv lo) (x - (bitSize lo))))
-            else LargeKey (shift lo x)
-                          (shift hi x .|. (conv $ shift lo (x - (bitSize lo))))
-         where conv = integerToLargeWord . largeWordToInteger
+         if x >= 0
+            then
+               LargeKey (shift lo x)
+                        (shift hi x .|. (convab $ shift lo (x - (bitSize lo))))
+            else
+               LargeKey (shift lo x .|. (convba $ shift hi (x + (bitSize hi))))
+                        (shift hi x)
+         where convab = integerToLargeWord . largeWordToInteger
+               convba = integerToLargeWord . largeWordToInteger
       largeBitSize ~(LargeKey lo hi) = largeBitSize lo + largeBitSize hi
 
-instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) => Show (LargeKey a b) where
+instance (Ord a, Bits a, Num a, LargeWord a, Bits b, Num b, LargeWord b) => Show (LargeKey a b) where
    showsPrec p = showInt . largeWordToInteger
 
-instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) => 
+instance (Ord b, Ord a, Bits a, Num a, LargeWord a, Bits b, Num b, LargeWord b) => 
    Num (LargeKey a b) where
       (+) = largeWordPlus
+      (-) = largeWordMinus
+      (*) a b =  go 0 0
+        where
+        go i r
+         | i == bitSize r = r
+         | testBit b i = go (i+1) (r + (a `shiftL` i))
+         | otherwise   = go (i+1) r
+      negate = id
+      abs    = id
+      signum a = if a > 0 then 1 else 0
       fromInteger = integerToLargeWord 
 
 -- Larger keys are instances of Bits provided their constituents are keys.
 
-instance (Ord a, Bits a, LargeWord a, Bits b, LargeWord b) => 
+instance (Ord a, Ord b, Bits a, Num a, LargeWord a, Bits b, Num b, LargeWord b) => 
    Bits (LargeKey a b) where
       (.&.) = largeWordAnd
       (.|.) = largeWordOr
       xor = largeWordXor
       shift = largeWordShift
+      complement (LargeKey a b) = LargeKey (complement a) (complement b)
       bitSize = largeBitSize
+      isSigned _ = False
+#if MIN_VERSION_base(4,6,0)
+      bit = bitDefault
+      testBit = testBitDefault
+      popCount = popCountDefault
+#endif
 
 instance (Ord a, Bits a, Bounded a, Integral a, LargeWord a, 
                  Bits b, Bounded b, Integral b, LargeWord b) => 
@@ -168,14 +198,36 @@ aoflk = undefined
 boflk :: (LargeKey a b) -> b
 boflk = undefined
 
-instance (Ord a, Bits a, LargeWord a, Ord b, Bits b, LargeWord b) =>
+instance (Bounded a, Bounded b, Enum b, Enum a, Ord a, Bits a, Num a, LargeWord a, Ord b, Bits b, Num b, LargeWord b) =>
    Integral (LargeKey a b) where
       toInteger = largeWordToInteger
+      quotRem a b =
+              let r = a - q*b
+                  q = go 0 (bitSize a) 0
+              in (q,r)
+       where
+       -- Trivial long division
+       go t 0 v = if v >= b then t+1 else t
+       go t i v
+              | v >= b    = go (setBit t i) i' v2
+              | otherwise = go t i' v1
+         where i' = i - 1
+               newBit = if (testBit a i') then 1 else 0
+               v1 = (v `shiftL` 1) .|. newBit
+               v2 = ((v - b) `shiftL` 1) .|. newBit
+      divMod = quotRem
 
-instance (Ord a, Bits a, LargeWord a, Ord b, Bits b, LargeWord b) =>
-   Real (LargeKey a b)
+instance (Ord a, Bits a, Num a, Bounded a, Bounded b, Enum a, Enum b, LargeWord a, Ord b, Bits b, Num b, LargeWord b) => Real (LargeKey a b) where
+      toRational w = toRational (fromIntegral w :: Integer)
 
-instance Enum (LargeKey a b)
+
+instance (Eq a, Bounded a, Num a, Enum b, Enum a, Bounded b, Num b) => Enum (LargeKey a b) where
+	toEnum i = LargeKey (toEnum i) 0
+	fromEnum (LargeKey l _) = fromEnum l
+	pred (LargeKey 0 h) = LargeKey maxBound (pred h)
+	pred (LargeKey l h) = LargeKey (pred l) h
+	succ (LargeKey l h) = if l == maxBound then LargeKey 0 (succ h)
+                                               else LargeKey (succ l) h
 
 type Word96  = LargeKey Word32 Word64
 type Word128 = LargeKey Word64 Word64
